/* 
   HTTP-DAV client routines.
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* FIXME: Ugly: This is only compiled in if USE_DAV is defined.
 * autoconf might make this nicer */
/* Set http_persist = true for persistent connections.
 * TODO: Persistent connections probably don't work yet. Probably. 
 */
#ifdef USE_DAV

#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "protocol.h"
#include "httpdav.h"
#include "common.h"
#include "socket.h"

/**** Public Interface ****/

/* Authorization method */
enum http_auth_t http_auth = http_auth_none;
bool http_persistent;

/**** Private Details ***/

/* The socked connected to the server */
int http_sock;

/* For storing username/password for when the request is made */
char *req_username, *req_password;
char *http_auth_basic_cookie; /* for holding the Base64'ized username:pass */
char *http_useragent = "sitecopy/" VERSION;

/* The local hostname, for the Host: header.
 * This is filled when dav_connect is called. */
char http_localhost[BUFSIZ];

/* We need to remember the remote host and port even after connecting
 * the socket so we can form the Destination: header */
char *remotehost;
struct in_addr remoteaddr;
int remoteport;

/* Time in seconds to wait for the server to give us a 100 Continue
 * after submitting a PUT with an "Expect: 100-continue" header.
 * This doesn't really matter too much */
#define CONT_TIMEOUT 5
int http_use_expect; /* == 0 if it we don't know, 1 if it does,
			    -1 if it doesn't support 100-continue */

/* For persistent connections, this is set to false when the socket
 * is closed by the server, so the next request has to open it again.
 * For non-persistent connections, this is not used */
bool http_connected;

/* The transfer encoding types */
enum http_te_t {
    /* TODO: Support other types */
    http_te_chunked,
    http_te_unknown
};

struct proto_driver dav_driver = {
    dav_connect, /* Connect */
    dav_userpass, /* Login */
    dav_disconnect, /* Disconnect */
    dav_move, /* File move */
    http_put, /* File upload - HTTP PUT */
    http_delete, /* File delete */
    NULL, /* File chmod */
    dav_mkcol, /* Dir create */
    dav_rmdir, /* Dir remove */
    NULL, /* Error string */
    "http", /* Service name */
    "HTTP-DAV" /* User-visible protocol name */
};

/* Makes an HTTP request to the server, using the given method,
 * to the given URI, including the specified headers, if given.
 * Give extra headers in form: 
 *   "Header: Value\nHeader2: Value2\n". Do not include two
 * trailing \n's.
 * Returns: PROTO_OK on success. */

int http_request( const char *method, const char *method_uri,
		  const char *headers ) {
    char buffer[BUFSIZ];
    int ret;

    snprintf( buffer, BUFSIZ, 
	      "%s %s HTTP/1.1\n"
	      "Host: %s\n"
	      "User-Agent: %s\n", 
	      method, method_uri, http_localhost, http_useragent );
    /* Could do with some overflow checks here */
    if( http_auth==http_auth_basic ) {
	strcat( buffer, "Authorization: Basic " );
	strcat( buffer, http_auth_basic_cookie );
	strcat( buffer, "\n" );
    }
    /* and here */
    if( headers!=NULL ) {
	strcat( buffer, headers );
    }
    strcat( buffer, "\n" );
    
    if( !http_persistent || (http_persistent && !http_connected) ) {
	if( (ret = dav_open() != PROTO_OK ) )
	    return ret;
    }

    debug( debug_some, ">>> HTTP Request:\n%s", buffer );
    /* FIXME: Check some return codes somewhere, here */

    send_data( http_sock, buffer );
    return PROTO_OK;
}

/* Simple HTTP put. 
 * local is the local filename. Remote is the destination URI (URI?)
 * Make it proper.
 * Returns:
 *   PROTO_FILE if no local file
 *   PROTO_ERROR if something general goes wrong
 *   PROTO_OK if it all works fine
 */
int http_put( const char *local, const char *remote ) {
    char headers[BUFSIZ];
    struct stat info;
    int ret;
    
    /* stat the local file... we need it's length, and to check it
     * exists before trying to send it.
     * FIXME: We should really fopen it then fstat it to prevent
     * anything silly happening between now and the send_file call */
    
    if( stat( local, &info ) == -1 )
	return PROTO_FILE;
	snprintf( headers, BUFSIZ,
		 "Content-Length: %d\n"
		 /* FIXME: What to send as content type? */
		 "Content-Type: text/html\n"
		 "Expect: 100-continue\n", info.st_size );
    
    http_request( "PUT", remote, headers );
    
    /* If we give a content-length of 0, Apache doesn't give 100-continue
	 * FIXME: Should it do this - refer to specs */
    if( info.st_size > 0 ) { 
	if( http_use_expect==0 || http_use_expect==1 ) {
	    /* Tricky bit - we don't want to send the whole file to the server 
	     * straight away, since it might just come back with an error, which
	     * would waste muchos bandwidth and time. So we use the clever HTTP/1.1
	     * Expect / 100-continue feature, which lets the server return an
	     * error code based on the headers or let us carry on before sending
	     * the body. 
	     * But, we have to accomodate for servers which DON'T respect
	     * the Expect: line. So, hang around for CONT_TIMEOUT seconds,
	     * waiting for anything to come down the wire.
	     * If nothing comes after that amount of time, just lob them
	     * the file. */
	    
	    fd_set sockfd;
	    struct timeval tv;
	    int resp;
	    /* Add the fd to the set */
	    FD_ZERO( &sockfd );
	    FD_SET( http_sock, &sockfd );
	    /* Set the timeout appropriately */
	    tv.tv_sec = CONT_TIMEOUT;
	    tv.tv_usec = 0;
	    /* And wait for something to happen */
	    resp = select( http_sock+1, &sockfd, NULL, NULL, &tv );
	    
	    switch( resp ) {
	    case -1: /* error */
		perror( "select" );
		return PROTO_ERROR;
	    case 0: /* timed out */
		http_use_expect = -1; /* don't try it again */
		break;
	    case 1: /* we got us a response! */
		http_use_expect = 1;
		if( http_getresponse( ) != 100 ) {
		    /* The server doesn't want the file. Bummer */
		    debug( debug_some, "Upload failed.\n" );
		    return PROTO_ERROR;
		} /* else... we got a valid response - send that file as normal */
		break;
	    default: /* eh-up lard? */
		debug( debug_some, "This really shouldn't have happened.\n" );
		return PROTO_ERROR;	    
	    }
	}
	
	/* Joy and happiness - the server DOES want the file */
	send_file( http_sock, local );
	
	switch( http_getresponse( ) ) {
	case 200: /* OK */
	case 201: /* Created */
	    debug( debug_some, "Upload successfull.\n" );
	    ret = PROTO_OK;
	    break;
	default:
	    debug( debug_some, "Upload failed.\n" );
	    ret = PROTO_ERROR;
	    break;
	}
    } else {
	/* Don't bother waiting for a response... */
	switch( http_getresponse( ) ) {
	case 201: /* Created */
	case 204: /* No content */
	    debug( debug_some, "Upload successfull.\n" );
	    ret = PROTO_OK;
	    break;
	default:
	    debug( debug_some, "Upload failed.\n" );
	    ret = PROTO_ERROR;
	    break;
	}
    }
    
    return ret;
}

/* This connects to the server.
 * Returns
 *  PROTO_LOOKUP if the hostname of the server could not be resolved
 *  PROTO_LOCALHOST if the hostname of the local machine could not be
 *    found
 *  PROTO_CONNECT if the socket could not be connected
 *  PROTO_OK if the connection was made successfully.
 */
int dav_connect( const char *hostname, const int port ) {
    /* Look up the remote host name, and the local host name, since
     * we need that for the HTTP/1.1 Host: request header */
    remotehost = strdup( hostname );
    remoteport = port;
    if( host_lookup( remotehost, &remoteaddr ) )
	return PROTO_LOOKUP;
    if( gethostname( http_localhost, BUFSIZ ) < 0 )
	return PROTO_LOCALHOST;
    http_persistent = true;
    debug( debug_some, "Local host is: %s\n", http_localhost );
    http_connected = false;
    http_use_expect = 0;
    return PROTO_OK;
}

/* Supply the user-specified login name and password */
int dav_userpass( const char *username, const char *password ) {
    char *temp;
    req_username = strdup( username );
    req_password = strdup( password );
    /* HTTP Basic Authentication:
     * Put "username:password" into temp, then base64 it */
    temp = malloc( strlen(username) + strlen(password) + 2 );
    strcpy( temp, username );
    strcat( temp, ":" );
    strcat( temp, password );
    http_auth_basic_cookie = base64( temp );
    http_auth = http_auth_basic;
    free( temp );
    return PROTO_OK;
}

    /* Close the connection */
int dav_disconnect( void ) {
    if( http_persistent && http_connected ) {
	socket_close( http_sock );
    }
    return PROTO_OK;
}

/* Gets the response from the server.
 * TODO: Read the headers, get the body.
 * Then persistent connections might work. */
int http_getresponse( void ) {
    char buffer[BUFSIZ], *part, *name, *value, *chunk;
    int status, body_length, chunk_size;
    enum http_te_t encoding;
    bool close_connection;

    if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
	return PROTO_ERROR;
    }
    /* FIXME: Check the spec for what this line should be */
    debug( debug_some, "HTTP response line: %s", buffer );
    if( strlen( buffer ) < 14 )
	/* We need at least "HTTP/1.1 xxx..." */
	return PROTO_ERROR;
    /* Parse the status code */
    part = strtok( buffer, " " ); /* "HTTP/1.1" */
    part = strtok( NULL, " " ); /* xxx */
    status = atoi( part );
    debug( debug_some, "HTTP reponse code: %d\n", status );

    /* Initialize body length to -1 to indicate we don't know what it is */
    body_length = -1;
    encoding = http_te_unknown;
    close_connection = false;

    /* Now read the rest of the header... up to the next blank line */
    while( read_line( http_sock, buffer, BUFSIZ ) > 0 ) {
	/* FIXME: What to check for here? */
	if( strcmp( buffer, "\r\n" ) == 0 ||
	    strcmp( buffer, "\n" ) == 0 ) {
	    debug( debug_some, "End of headers.\n" );
	    break;
	}
	debug( debug_some, "< [%d] %s", strlen( buffer ), buffer );
	/* Now parse the header line. This is all a bit noddy. */
	name = strtok( buffer, ":" );
	if( name != NULL ) {
	    value = strtok( NULL, ":" );
	    /* Strip leading whitespace, and trailing CRLF */
	    while( *value!='\0' && *value==' ' ) value++;
	    if( (part = strchr( value, '\r' )) != NULL ) 
		*part = '\0';
	    if( (part = strchr( value, '\n' )) != NULL ) 
		*part = '\0';
	    debug( debug_some, "Header: [%s] = [%s]\n", name, value );
	    if( strcasecmp( name, "Content-Length" ) == 0 ) {
		body_length = atoi( value );
	    } else if( strcasecmp( name, "Transfer-Encoding" ) == 0 ) {
		if( strcasecmp( value, "chunked" ) == 0 ) {
		    encoding = http_te_chunked;
		} 
	    } else if( strcasecmp( name, "Connection" ) == 0 ) {
		if( strcasecmp( value, "close" ) == 0 ) {
		    close_connection = true;
		    debug( debug_some, "Non-persistent connection!\n" );
		}
	    }
	}
    }
    if( body_length==-1 && encoding==http_te_unknown ) {
	/* Then we don't know how to get the body. Oh dear, give up. */
	debug( debug_some, "Can't get body.\n" );
    } else if( encoding==http_te_chunked ) {
	/* Get the body in chunked encoding.
	 * HTTP/1.1 draft 05 tells how to do this... */
	do {
	    if( read_line( http_sock, buffer, BUFSIZ ) < 0 )
		return -1;
/*	    debug( debug_full, "< %s\n", buffer ); */
	    sscanf( buffer, "%x", &chunk_size );
	    debug( debug_full, "Got chunk size: %d\n", chunk_size );
	    if( chunk_size > 0 ) {
		chunk = malloc( chunk_size );
		/* TODO: Make this decent. */
		read( http_sock, chunk, chunk_size );
		debug( debug_full, "<<< Chunk:\n%s\n", chunk );
		free( chunk );
		/* Skip CRLF */
		if( read_line( http_sock, buffer, BUFSIZ ) < 0 )
		    return -1;
	    }
	} while( chunk_size > 0 );
	/* Read in that last line */
	if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
	    /* It broke, kill it */
	    close_connection = true;
	}
    } else {
	/* No encoding, but we gotta have a Content-Length.
	 * TODO: Erm, will this work? */
	while( body_length > 0 ) {
	    if( (chunk_size = read( http_sock, buffer, BUFSIZ )) < 0 )
		return -1;
	    body_length -= chunk_size;
	}
    }
    debug( debug_full, "Persist: %d, close: %d\n", http_persistent,
	  close_connection );

    /* Now, close the socket? */
    if( !http_persistent || close_connection ) {
	dav_close( );
    }
    return status;
}

/* Perform the file operations */
int dav_move( const char *from, const char *to ) {
    char headers[BUFSIZ];
    int ret;

    snprintf( headers, BUFSIZ, "Destination: http://%s:%d%s\n", 
	      remotehost, remoteport, to );

    http_request( "MOVE", from, headers );

    switch( http_getresponse( ) ) {
    case 201: /* created new dest */
    case 204: /* moved over existing dest */
	ret = PROTO_OK;
	break;
    default:
	ret = PROTO_ERROR;
	break;
    }
    return ret;
}

/* Deletes the specified resource on the server */
int http_delete( const char *filename ) {
    int ret;

    http_request( "DELETE", filename, NULL );

    switch( http_getresponse( ) ) {
    case 204: /* successfull delete? */
	ret = PROTO_OK;
	break;
    default:
	ret = PROTO_ERROR;
	break;
    }	

    return ret;
}

/* Delete the specified collection on the server.
 * This is separate to get rid of the / */
int dav_rmdir( const char *dirname ) {
    int ret;
    char *filename;
    
    filename = malloc( strlen(dirname) + 2 );
    strcpy( filename, dirname );
    strcat( filename, "/" );

    http_request( "DELETE", dirname, NULL );

    switch( http_getresponse( ) ) {
    case 204: /* No content */
	ret = PROTO_OK;
	break;
    default:
	ret = PROTO_ERROR;
	break;
    }	

    return ret;
}


int dav_mkcol( const char *dirname ) {
    int ret;
    char *uri;
    
    uri = malloc( strlen(dirname) + 2 );
    strcpy( uri, dirname );
    strcat( uri, "/" );

    http_request( "MKCOL", dirname, NULL );

    switch( http_getresponse( ) ) {
    case 201: /* successfull create */
	ret = PROTO_OK;
	break;
    default:
	ret = PROTO_ERROR;
	break;
    }

    return ret;
}

int dav_open( void ) {
    debug( debug_some, "Connecting to %s:%d...\n", remotehost, remoteport );
    if( ( http_sock = socket_connect( remoteaddr, remoteport ) ) < 0 )
	return PROTO_CONNECT;
    debug( debug_some, "Connected.\n" );
    http_connected = true;
    return PROTO_OK;
}

int dav_close( void ) {
    debug( debug_some, "Closing socket.\n" );
    socket_close( http_sock );
    http_connected = false;
    debug( debug_some, "Socket closed.\n" );
    return PROTO_OK;
}    

/*    int (*file_chmod) ( const char *filename, const mode_t mode );

    int (*dir_remove) ( const char *dirname );
*/

#endif /* USE_DAV */
